#include <iostream>
#include <fstream>
#include <string>
#include <map>
#include <stdlib.h>
#include <time.h>
#include <cstdio>

#include "shaunlib.h"
#include "driver.h"
#include "whap.h"
#include "haplotype.h"
#include "reader.h"
#include "parameters.h"
#include "helper.h"
#include "window.h"

using namespace mymath;
using namespace std;

void displayPhase(CPhase &);

const string VERSION = "v2.06";
const string DATE    = "13/Dec/04";
const string AUTHOR  = "S. Purcell, P. Sham" ;
const string CONTACT = "spurcell@pngu.mgh.harvard.edu";


/////////////////////////////////////
// Driver: 
//    -process command line
//    -read genotype data
//    -invoke SNPHAP to generate haplotypes
//    -call whap procedure

int main(int argc, char* argv[])
{

    CRandom::srand(time(0));
    cout.setf(ios::fixed);
    cout.precision(3);
    
    
    ////////////////////////////////////////////
    // Read and process command line arguments 

    // Phenotype and haplotype data
    CSample sample; 
    CDatfile dat;
    CMap locmap;

    // Main driver class
    Driver drive(sample,dat,locmap);

    // Command line args
    CArgs a(argc,argv);
    drive.a = &a;

    // Process command line
    drive.firstPass();

    // title
    if (!par::nice_dump)
	cout << "\nWHAP! | "
	     << VERSION << " | " 
	     << DATE << " | "
	     << AUTHOR << " | " 
	     << CONTACT << "\n";


    // Initiate sliding Window
    drive.initWindow();
    
    // Store non-permuted original 
    drive.storeOriginal();
    
    // Pre-permute datasets?
    if (par::prepermute && par::tperm)
	drive.prePermute();

    if (par::prepermute && par::gperm)
	drive.prePermuteW();


    // Some summary statistics regarding trait and sample
    helper::summary(sample);
    
    // Run Whap
    if (par::window)
	while (1)
	{
	    drive.slideWindow_BASIC();
	    drive.initWhap();
	}
    else
	drive.initWhap();	
    
    exit(0);
}




void Driver::firstPass()
{

    // Process command line arguments

    if (a->find("--verbose")) par::verbose = true;
    if (a->find("--dump")) par::dump = true;
    if (a->find("--phase")) 
    { 
	par::nice_dump = true; 
	par::collapse_phase_speedup = false; 
    }
    if (a->find("--debug")) par::debug = true;
    if (a->find("--keep")) par::keep_snphap_files = true;
    if (a->find("--noweight")) par::weight = false;
    if (a->find("--constrain")) par::usr_constraint = true;
    if (a->find("--xchr")) par::xchr = true;
    if (a->find("--dom")) par::model_dom = true;
    if (a->find("--rec")) par::model_rec = true;
    if (a->find("--joint")) par::joint = true;
    if (a->find("--ci")) 
    { 
	par::ci = true; 
	par::civ = atof(a->value("--ci").c_str()); 
    } 
    if (a->find("--usat")) { par::usat = true; }
    if (a->find("--cond")) par::cond = true;
//    if (a->find("--recomb")) 
//	par::recomb = atoi(a->value("--recomb").c_str());
    if (a->find("--allhaps")) par::calcfullhaps = true;
//    if (a->find("--haps")) par::readhaps = true;
    if (a->find("--largest")) par::hap_specific = true;
    if (a->find("--hs") || a->find("--hap-specific")) par::hap_specific = true;
    if (a->find("--hs")) par::hs_min_output = true;
    if (a->find("--wald")) par::wald = true;
    if (a->find("--sec") || a->find("--second")) par::secondary = true;
    if (a->find("--sll")) { par::secondary = true; par::sec_global = false; } 
    if (a->find("--sverbose")) par::secondary_verbose = true;
    if (a->find("--window")) par::window = true;
    if (a->find("--winstep")) 
	par::winstep = atoi(a->value("--winstep").c_str());
    if (a->find("--gap")) par::bp_gap = atoi(a->value("--gap").c_str());
    if (a->find("--allres")) par::win_all_results = true;
    if (a->find("--th")) par::threshold = atof(a->value("--th").c_str());
    if (a->find("--at")) 
	par::abs_threshold = atof(a->value("--at").c_str())/100; 
    if (a->find("--ap")) par::abs_threshold = atof(a->value("--ap").c_str()); 
    if (a->find("--cov")) par::alt_cov = par::null_cov = true;
    if (a->find("--alt-cov")) par::alt_cov  = true;
    if (a->find("--null-cov")) par::null_cov = true;
    if (a->find("--gxe")) 
	par::alt_gxe = par::null_gxe = par::alt_cov = par::null_cov = true;
    if (a->find("--alt-gxe")) par::alt_gxe = par::alt_cov = true;
    if (a->find("--null-gxe")) par::null_gxe = par::null_cov = true;
    if (a->find("--gxe0")) par::alt_gxe = par::null_gxe = true;
    if (a->find("--alt-gxe0")) par::alt_gxe = true;
    if (a->find("--null-gxe0")) par::null_gxe = true;
//    if (a->find("--alt-po")) par::alt_po = true;
//    if (a->find("--po")) par::alt_po = par::null_po = true;
    
    if (a->find("--perm")) 
    { 
	par::perm = atoi(a->value("--perm").c_str());
	par::tperm = true;
    }
    if (a->find("--wperm")) 
    { 
	par::perm = atoi(a->value("--wperm").c_str()); 
	par::gperm = true; 
    }

    // note: currently, do not distinguish between --1+ and --1-
    if (a->find("--1")) par::one_sided_pos = true;
    if (a->find("--1+")) par::one_sided_pos = true;
    if (a->find("--1-")) par::one_sided_neg = true;
    if (par::one_sided_pos && par::one_sided_neg)
      {
	cerr << "Using a two-sided test: you set both --1+ and --1-";
	par::one_sided_pos = par::one_sided_neg = false;
      }
    
    if (a->find("--no_preperm")) 
	par::prepermute = false;
    if (a->find("--eps")) 
	par::eps = atof(a->value("--eps").c_str());
    if (a->find("--tol")) 
	par::tol = atof(a->value("--tol").c_str());
    if (a->find("--iter")) 
	par::mxiter = atoi(a->value("--iter").c_str());
    if (a->find("--SAtemp")) 
	par::sa_temp = atoi(a->value("--SAtemp").c_str());
    if (a->find("--SAcycle")) 
	par::sa_cycle = atoi(a->value("--SAcycle").c_str());
    if (a->find("--SAdecline")) 
	par::sa_decline = atof(a->value("--SAdecline").c_str());
    if (a->find("--SAiter")) 
	par::sa_mxiter = atoi(a->value("--SAiter").c_str());
    if (a->find("--repeat")) 
	par::sa_repeat = atoi(a->value("--repeat").c_str());

    if (a->find("--missing-pheno")) 
	CSample::missing_pheno = atoi(a->value("--missing-pheno").c_str());
    if (a->find("--missing-geno")) 
	CSample::missing_geno = atoi(a->value("--missing-geno").c_str());

//    if (a->find("--pp")) // not yet fully implemented...
//	par::model_parental_phenos = true;

    if (a->find("--model"))
    {
	par::alt_between = par::alt_within = par::alt_total 
	    = par::null_between = par::null_within = par::null_total = false;

	string mod = a->value("--model");
	// both alt and null specified? (if '/' in mod)
	string alt_mod, null_mod;
	if ( mod.find("/",0) != string::npos ) 
	{
	    alt_mod = mod.substr(0, mod.find("/",0));
	    null_mod = mod.substr(mod.find("/",0)+1,
				  mod.length()-1-mod.find("/",0)); 	    
	} else 	{ alt_mod = mod; null_mod = "0"; }
	// process
	if ( alt_mod.find("b",0) != string::npos 
	     || alt_mod.find("B",0) != string::npos ) 
	    par::alt_between = true;
	if ( alt_mod.find("w",0) != string::npos 
	     || alt_mod.find("W",0) != string::npos ) 
	    par::alt_within = true;
	if ( alt_mod.find("t",0) != string::npos 
	     || alt_mod.find("T",0) != string::npos ) 
	    par::alt_total = true;
	if ( null_mod == "0" ) {
	    par::null_between = par::alt_between;
	    par::null_within = par::alt_within;
	    par::null_total = par::alt_total;
	} else {
	    if ( null_mod.find("b",0) != string::npos 
		 || null_mod.find("B",0) != string::npos ) 
		par::null_between = true;
	    if ( null_mod.find("w",0) != string::npos 
		 || null_mod.find("W",0) != string::npos ) 
		par::null_within = true;
	    if ( null_mod.find("t",0) != string::npos 
		 || null_mod.find("T",0) != string::npos ) 
		par::null_total = true;
	}
	if (!par::alt_between && !par::alt_within 
	    && !par::alt_total) error("Nonsensical --model for alternate");
    }

    if (a->find("--mean"))
    {
	par::fix_mean = true;
	par::fixed_mean = atof(a->value("--mean").c_str());
    }

    if (a->find("--var"))
    {
	par::fix_var = true;
	par::fixed_var = atof(a->value("--var").c_str());
    }

    if (a->find("--prev"))
    {
	par::fix_mean = true;
	par::fixed_mean = atof(a->value("--prev").c_str());
	par::fixed_mean = log(par::fixed_mean / (1-par::fixed_mean));
    }
    
    if (a->find("--no_collapse")) par::collapse_phase_speedup = false;

    if (a->find("--uninform")) 
	par::uninform_threshold = atoi(a->value("--uninform").c_str());	
    
    if (a->find("--no_trio_b")) par::no_between_trios = true;


    if (a->find("--help") || a->find("-h") || a->count() == 0 ) 
    {
	cout << "whap --file {root} [--verbose] \n"
	     << "                   [--dump] \n"
	     << "                   [--alt 1,2,3,4] \n"
             << "                   [--null 2,3] \n"
	     << "                   [--model bwf/bwf] \n"
	     << "                   [--constrain 1,1,1,2/1,1,1,1] \n"
	     << "                   [--mean 0] \n"
	     << "                   [--var 1] \n"
	     << "                   [--prev 0.1] \n"
	     << "                   [--joint] \n"
	     << "                   [--cond] \n"
	     << "                   [--window] \n"
	     << "                   [--winstep 1] \n"
	     << "                   [--allres] \n"
	     << "                   [--perm 10000 ] \n"
	     << "                   [--preperm] \n"
	     << "                   [--wperm 10000 ] \n"
	     << "                   [--eps 1e-8] \n"
	     << "                   [--tol 1e-6] \n"
	     << "                   [--iter 20000] \n"
	     << "                   [--SAtemp 10] \n"
	     << "                   [--SAcycle 10] \n"
	     << "                   [--SAdecline 0.8] \n"
	     << "                   [--SAiter 25] \n"
	     << "                   [--repeat 1] \n"
	     << "                   [--noweight] \n"
	     << "                   [--allhaps] \n"
	     << "                   [--haps] \n"
	     << "                   {--ignore-parents} \n"
	     << "                   [--wald] \n"
	     << "                   [--sec] \n"
	     << "                   [--sverbose] \n"
	     << "                   [--th 2] \n"
	     << "                   [--at 2] \n"
	     << "                   [--cov] \n"
	     << "                   [--missing-pheno -9] \n"
	     << "                   [--missing-geno 0] \n"
	     << "                   [--help]\n"
	     << "  where root.ped is a pedigree file\n"
	     << "        root.dat is a descriptive data file\n"
	     << "        root.map is a map file\n";
	exit(0);
    }


    string filename;
    if (!a->find("--file")) error("Must specify --file : use --help");
    filename = a->value("--file");

    if (par::hap_specific && par::secondary)
	error("Can only specify --sec in omnibus test mode (i.e. not --largest)");

    if ( (par::model_dom || par::model_rec) && !par::hap_specific)
      error("Can only specify dominant or recessive models for --hap-specific tests");
    
    

    // WARNINGS...

    if ( par::model_dom && par::model_rec ) 
      {
	cout << "** <<Warning>> set both dominant and recessive models\n"
	     << "               changing to standard multiplicative model\n";
	par::model_dom = par::model_rec = false;
      }

    if (par::secondary && par::calcfullhaps)
	cout << "** <<Warning>> secondary analysis only checked "
	     << "for when --allhaps option is not in effect ** \n";
    if (par::cond && !par::gperm)
	cout << "** <<Warning>> use the --wperm N option if "
	     << "analysing affected only TDT samples \n";
    
    if (par::null_cov && !par::alt_cov) 
	error("Covariates cannot be specified under the null if not also under the alternate.");
    if (par::null_gxe && !par::alt_gxe) 
	error("Moderators cannot be specified under the null if not also under the alternate.");

    
    ////////////////////////////////////////////
    // Read and process datafile 
    
    // Read data files in this order : map, dat, ped
    dat.readDatfile(filename+".dat");
    locmap.readMapfile(filename+".map",dat);
    sample.readPedfile(filename+".ped",dat,locmap);


    // look for usat details if specified
    if (par::debug) cerr << "About to downcode usats, if any\n";
    if (par::usat) helper::usatFile(filename+".all", 
				    acount,
				    downcode,
				    sample.countLocus());

    // assign a 'box' of CBox to each individual
    // (via handle class)
    for (int i=1; i<=sample.countFamily(); i++)
	for (int j=1; j<=sample.fam(i)->nInFamily(); j++)
	    sample.fam(i)->ind(j)->box.set(new CBox(sample.fam(i)->ind(j)));


    // Number of individuals / trios in sample
    int n = sample.countFamily();

    // Number of covariates in samples
    if (par::alt_cov || par::alt_gxe) 
	par::covn = sample.countCovariate();

    // Recode from 1,2,3,4 -> 1,2 for each locus
    if (par::debug) cerr << "About to recode SNPs 1/2 coding\n";
    helper::recodeSNPs(sample);

    // Check if a quantitative trait
    // Currently, no missing phenotypes allowed... 
    if (sample.countQTrait() > 0) par::binary = false;
    else if (sample.countBTrait() > 0) par::binary = true;
    else error("No phenotypes (binary or continuous) specified");

    // Flag if (both) parents are present 
    if (par::debug) cerr << "About to determine presence of any unrelateds/trios\n";
    for (int i=1; i<=sample.countFamily(); i++)
    {
	// 'parents' is false by default

	CFamily * family = sample.fam(i);
	// a trio? 
	if (family->nInFamily() == 3)
	{
	    CIndividual * kid = &family->getChildFromTrio();
	    if (kid == NULL) error("Problem with family structure");
	    kid->box->parents = true;	    
	    family->proband = kid;
	    par::no_trios = false;
	}
	else if (family->nInFamily() == 1) 
	{
	    family->proband = family->ind(1);
	    par::no_unrel = false;
	}
	else error("Not TRIO or UNRELATED");	    
    }

    // Okay to model parental quantitative traits?
    if (par::model_parental_phenos)
    {
	if (par::no_trios) error("Cannot use --pp: No trio data in sample");
	if (par::binary) error("Cannot use --pp: only good for quantitative traits");
    }

   
    // Calculate sample variance for trait and any covariates
    helper::sampleVariances(sample);

    // can we set the speedup for qualitative conditional analysis?
    if (par::cond && par::binary && !par::alt_cov && !par::alt_gxe) 
    {
	par::qual_cond_speedup = true;
//	cout << "-- qualcond speedup in effect -- \n";
    }
    ///////////////////////////////////////////
    // Preliminary parse of --alt and --null command line 
    if (par::debug) cerr << "About to prelimanary parse of --alt\n";
    if (a->find("--alt") && !par::calcfullhaps)
    {
	string s = a->value("--alt");
	string t;
	
	for (int i=0 ; i < s.length() ; i++)
	    if (s[i] == ',' || i == s.length()-1 ) 
	    {
		if (i == s.length()-1) t += s[i];
		alt.push_back(atoi(t.c_str()));
		t = "";
	    }
	    else
		t += s[i];
    }
    else if (par::window)
    {
	// If --window in effect but no --alt given, 
	// then assume --alt 1
	alt.resize(1);
	alt[0] = 1;
    }
    else
    {
	// Either no --alt
	// or --allhaps given
	// and no --window option

	// get number of loci 
	// (half cols in dat, minus ID, pheno, cov)
	int n;
	if (!par::usat) n=sample.countLocus();
	else n=acount.size();
	alt.resize(0);
	for (int i=1; i<=n; i++)
	    alt.push_back(i);
    }
    
    // Read null markers
    if (par::debug) cerr << "About to prelimanary parse of --null\n";
    if (a->find("--null"))
    {
	string s = a->value("--null");
	string t;
	
	for (int i=0 ; i < s.length() ; i++)
	    if (s[i] == ',' || i == s.length()-1 ) 
	    {
		if (i == s.length()-1) t += s[i];
		null.push_back(atoi(t.c_str()));
		t = "";
	    }
	    else
		t += s[i];
    }
    else
	null.resize(0); // set for 0 permanently
 

    
    ///////////////////////////////////
    // Parse user-specified constaints

    if (par::usr_constraint)
    {
	// makes no sense to also specify --null 
	// as the --constrain must do this
	if (a->find("--null")) error("Do not specify --null and --constrain together...");
	
	// to check: break if --allhaps --alt specified ?
	
	// constraint list should be comma-delimited and alt and null
	// are separated by a forward slash
	string con = a->value("--constrain");
	// both alt and null specified? (if '/' in mod)
	string alt_con, null_con;
	if ( con.find("/",0) != string::npos ) 
	{
	    alt_con = con.substr(0, con.find("/",0));
	    null_con = con.substr(con.find("/",0)+1,
				  con.length()-1-con.find("/",0)); 	    
	} else 	
	{ 
	    error("Bad --constrain list: should take form --constain 1,1,2/1,1,1 ");
	}	    
	
	// process alt constraint list
	string t;
	for (int i=0 ; i < alt_con.length() ; i++)
	    if (alt_con[i] == ',' || i == alt_con.length()-1 ) 
	    {
		if (i == alt_con.length()-1) t += alt_con[i];
		usr_alt.push_back(atoi(t.c_str()));
		t = "";
	    }
	    else
		t += alt_con[i];
	
	// and for null
	for (int i=0 ; i < null_con.length() ; i++)
	    if (null_con[i] == ',' || i == null_con.length()-1 ) 
	    {
		if (i == null_con.length()-1) t += null_con[i];
		usr_null.push_back(atoi(t.c_str()));
		t = "";
	    }
	    else
		t += null_con[i];
    	
	
	// number of haplotypes must match for alt and null
	if (usr_alt.size() != usr_null.size())
	    error("Bad --constain list: number of alt/null haps must match.");
	
	// ASSUME: sequential coding 1,2,3,4, etc
	// ASSUME: number of haps specified corresponds to the number in analysis
	//         after any --alt and exclusion of rare haplotypes, etc. 
	
    }
    
}



void Driver::initWindow()
{
    
    ////////////////////////////////////////////
    // Set drive.win.alt and drive.win.null

   // Standard analysis is a subset of a sliding window
    if (!par::window)
    {
	// Create a single Window 
	win.push_back(new Window);
	// Interpret marker list directly
	// i.e. not a template
	win[0]->alt = alt;
	win[0]->null = null;
	win[0]->pos = 1;
    }
    else
    {
	// Given template such as --alt 1,3,5,7 --null 1,7
	// Template must include 1
	
	// Get total number of loci
	total_nsnps= sample.countLocus();
	if (!par::usat) total_nloci = total_nsnps;
	else total_nloci = acount.size();
	
	// Check that --alt contains 1 
	// (otherwise it doesn't make sense as a template)
	bool match = false;
	for (int i=0; i<alt.size(); i++)
	    if (alt[i] == 1) match = true;
	if (!match) error("--alt must contain 1 with --window");
	
	// get length of --alt window; min must eq 1 (static)
	Window::size = alt[0];
	for (int i=1; i<alt.size(); i++)
	    if (alt[i] > Window::size) Window::size = alt[i];

	// set finishing point for sliding window (static)
	Window::end = total_nloci - Window::size + 1;

	// set starting position (always -1)
	Window::n = -1;

	// determine vector of gaps (defined by par::gap_bp)
	// if a gap between LOCUS 6 and 7, store 6
	if (!par::usat)
	{
	    for (int i=1; i < total_nloci ; i++)
		if ( locmap.bp[sample.loc(i+1)->name] 
		     - locmap.bp[sample.loc(i)->name] > par::bp_gap )
		    Window::gap.push_back(i);
	}
	else
	{
	    int n=1;
	    for (int i=0; i < acount.size()-1  ; i++)
	    {
		if ( locmap.bp[sample.loc(n+1)->name] 
		     - locmap.bp[sample.loc(n)->name] > par::bp_gap )
		    Window::gap.push_back(n);
		n += acount[i];
	    }
	}
	
	// Display to screen a list of gaps (numbers are LOCI, not SNPs)
	for (int i=0; i < Window::gap.size() ; i++)
	    cout << "\nSkip between marker " << Window::gap[i] << " and " << Window::gap[i]+1 
		 << " (" << ( locmap.bp[sample.loc(Window::gap[i]+1)->name] 
			      - locmap.bp[sample.loc(Window::gap[i])->name] ) / 1000 << " kb)"; 
	cout << "\n";
    }

}



void Driver::initWhap()
{

    ////////////////////////////////////
    // Create file to send to SNPHAP

    // get number of founders to be phased
    int founders = 0;
    for (int i=1; i<=sample.countFamily(); i++)
	for (int j=1; j<=sample.fam(i)->nInFamily(); j++)
	{
	    if (!par::xchr)
	    {
		if (sample.fam(i)->ind(j)->isFounder()) founders++;
	    }
	    else
	    {
		if (sample.fam(i)->ind(j)->isFounder() && sample.fam(i)->ind(j)->sex == 2) founders++;
	    }
	}

    // get number of SNPs to be sent (downcode uSats if necessary)
    int wsize = 0;
    if (!par::usat) wsize = win[Window::n]->alt.size();
    else for (int i=0; i<win[Window::n]->alt.size();i++) 
	wsize += downcode[win[Window::n]->alt[i]-1 ].size();

    // create temporary storage, for export to file
    CData d(founders,1+wsize*2);
    d.include_all();
    d.include_all_var();

    // store ID -> FID, IID map
    IntArray id1;
    StringArray id2;
    StringArray id3;
    id1().resize(founders);
    id2().resize(founders);
    id3().resize(founders);

    // setup coding system, so that markers can be tracked under the null
    // e.g. --alt 1,4,5 will appear as only 3 markers, 
    // so --null 4,5 would be invalid, otherwise: 1,4,5 stored in s_recode[]
    vector<int> s_recode(0);
    for (int k=0; k < win[Window::n]->alt.size(); k++)    
	s_recode.push_back(win[Window::n]->alt[k]);

    
    // Send only genotypes of founders
    int f=1, l=2;
    for (int i=1; i<=sample.countFamily(); i++)
	for (int j=1; j<=sample.fam(i)->nInFamily(); j++)
	{
	    if (sample.fam(i)->ind(j)->isFounder()) 
	    {
		if ( ( par::xchr && sample.fam(i)->ind(j)->sex == 2) || !par::xchr )
		{
		    CIndividual * person = sample.fam(i)->ind(j);
		    
		    l=2; // skip ID field
		    
		    // f is temporary ID -- map to the true Fam/Ind IDs
		    d[f][1] = f;
		    id1(f) = f;
		    id2(f) = person->family->getID();
		    id3(f) = person->getID();
		    
		    // create output file of genotypes 
		    if (!par::usat)
		    {
			for (int k=0; k < win[Window::n]->alt.size(); k++)    
			{
			    d[f][l++] = atoi(sample.fam(i)->ind(j)->loc(win[Window::n]->alt[k]).one.c_str());
			    d[f][l++] = atoi(sample.fam(i)->ind(j)->loc(win[Window::n]->alt[k]).two.c_str());
			}
		    }
		    else // have to handle usats differently...
		    {
			// if acount =  1   1   4       1   2    5   1 
			// and we ask for locus 3,4,5, this means SNPs
			//              1   2   3/4/5/6 7   8/9         <= stored in downcode...
			// => 3,4,5,6,7,8,9
			
			for (int k=0; k < win[Window::n]->alt.size(); k++)    
			{
			    for (int k1=0; k1<downcode[win[Window::n]->alt[k]-1].size(); k1++)
			    {
				d[f][l++] = atoi(sample.fam(i)->ind(j)->loc(downcode[win[Window::n]->alt[k]-1][k1]).one.c_str());
				d[f][l++] = atoi(sample.fam(i)->ind(j)->loc(downcode[win[Window::n]->alt[k]-1][k1]).two.c_str());
			    }
			}
		    }
			
		    f++;
		}
	    }
	}
    
    // write input file of founder genotypes for SNPHAP
    string out_name = uniqname("w-inp");
    d.write(out_name);
    
    if (par::debug) 
	cerr << "about to runsnphap()\n";
    
    // Run SNPHAP , populate sample O, K.
    // returns # of haplotypes found
    // also sets ind::nloci to number of loci
    int nhap = 0 ;
    nhap = runsnphap(sample,id1,id2,id3,*this,out_name);

    if (!par::keep_snphap_files)
	remove(out_name.c_str());

    if (par::debug) 
	cerr << "done runsnphap()\n";


    //////////////////////////////////////////////////
    // process model for ALTERNATE
    // alt_estimate() = T  -> keep in model
    // alt_estimate() = F  -> drop from model
    
    ind::alt_estimate().resize(ind::nloci);
    
    ////////////////////////////////////////////////
    // have to deal specially with --allhaps option
    
    if (par::calcfullhaps && a->find("--alt"))
    {
	// set to drop all
	for (int i=1; i<=ind::nloci;i++) ind::alt_estimate(i) = false;
	
	// need to re-read --alt list (as drive.alt[] would have been set to all)
	win[Window::n]->alt.resize(0);
	
	string s = a->value("--alt");
	string t;
	
	for (int i=0 ; i < s.length() ; i++)
	    if (s[i] == ',' || i == s.length()-1 ) 
	    {
		if (i == s.length()-1) t += s[i];
		win[Window::n]->alt.push_back(atoi(t.c_str()));
		t = "";
	    }
	    else
		t += s[i];
	
	// set markers to be kept 
	for (int i=0; i < win[Window::n]->alt.size(); i++)
	{
	    if (win[Window::n]->alt[i] <= 0 || win[Window::n]->alt[i] > ind::nloci)  
		error("Tried to specify a nonexistent locus in --alt");
	    ind::alt_estimate(win[Window::n]->alt[i]) = true;
	}
	
	// remake s_recode[]
	s_recode.resize(0);
	for (int i=0; i < win[Window::n]->alt.size(); i++)
	    s_recode.push_back(win[Window::n]->alt[i]);
	
    }
    else
    {
	// set all (already dropped before sending to SNPHAP)
	for (int i=1; i<=ind::nloci;i++) ind::alt_estimate(i) = true;
    }
    
    
    //////////////////////////////////////////////////
    // process submodel for NULL
    ind::null_estimate().resize(ind::nloci);
    
    // set to drop all
    for (int i=1; i<=ind::nloci;i++) ind::null_estimate(i) = false;

    // set markers to be kept
    for (int i=0; i < win[Window::n]->null.size(); i++)
    {

	// check in s_recode[j] 
	// if match, new code is j+1
	
	bool match = false;
	int loc = 0;
	for (int j=0; j < s_recode.size(); j++)
	    if ( win[Window::n]->null[i] == s_recode[j] ) 
	    {
		match = true;
		loc = j+1;    // new locus code
	    }
	
	if (!match) 
	    error("Tried to specify invalid locus in --null");
	
	ind::null_estimate(loc) = true;
	
    }
    
    
    ///////////////////////////////////////////////
    // Dump sample structure if --dump in effect
    if (par::dump)
    {
 	// First -- haplotype freqs
	
	map<CHaplotype, int>::iterator iter;
	for (iter = ind::hcode.begin(); iter != ind::hcode.end(); iter++)
 	{
	    int h= iter->first.length();
	    for (int n=1; n <= h; n++)
		cout << iter->first.pos(n);
	    cout << "\t" << " = " << ind::hapfreq1[iter->second] ; 
	    cout << "\n";
 	}
	
	cout << "\n----------------\n";
	
	// Iterate through families
	for (int i=1; i <= sample.countFamily(); i++)
	{
	    cout << "\n-------------------------------------";
	    
	    if (sample.fam(i)->nInFamily() > 1) 
	    {
		// Trio
		cout << "\nFamily " << sample.fam(i)->getID();
		CIndividual & person = sample.fam(i)->getChildFromTrio();
		cout << "\nIndividual " << person.family->getID() 
		     << "_" << person.getID();
		if (par::binary)
		    cout << "\nBinary trait " << person.bt(1);
		else
		    cout << "\nQuantitative trait " << person.qt(1);
		int n = person.box->PP.size();
		for (int p=0; p < n ; p++)
		{
		    cout << "\npat = " ; displayPhase(person.box->PP[p].father);
		    cout << "\nmat = " ; displayPhase(person.box->PP[p].mother);
		    for (int p2=0; p2 < person.box->PP[p].P.size(); p2++)
		    {
			cout << "\nkid phase : ";
			displayPhase(person.box->PP[p].P[p2]);
			
		    }
		}		
	    }
	    else
	    {
		// Individual
		cout << "\nUnrelated " << sample.fam(i)->getID();
		CIndividual & person = *sample.fam(i)->ind(1);
		cout << "\nIndividual " << person.family->getID() 
		     << "_" << person.getID();
		if (par::binary)
		    cout << "\nBinary trait " << person.bt(1);
		else
		    cout << "\nQuantitative trait " << person.qt(1);

		for (int p2=0; p2 < person.box->P.size(); p2++)
		{
		    cout << "\nind phase : ";
		    displayPhase(person.box->P[p2]);		    
		}

	    }
	    cout << "\n";
	    
	}
    
    }   



    //////////////////////////////////
    // Count up informative individuals

    // Iterate through families
    int inform_count = 0;
    for (int i=1; i <= sample.countFamily(); i++)
    {
	int indiv_inform = 0;
	// Trio
	if (sample.fam(i)->nInFamily() > 1) 
	{
	    CIndividual & person = sample.fam(i)->getChildFromTrio();
	    int n = person.box->PP.size();
	    for (int p=0; p < n ; p++)
		for (int p2=0; p2 < person.box->PP[p].P.size(); p2++)
		    indiv_inform++;
	}
	else
	{
	    // Individual
	    CIndividual & person = *sample.fam(i)->ind(1);
	    for (int p2=0; p2 < person.box->P.size(); p2++)
		indiv_inform++;
	}
	
	if (indiv_inform>0) inform_count++;
    }

    if (!par::nice_dump)
	cout << "\n" << inform_count << " of " << sample.countFamily() << " individuals/trios are informative";
    

    ////////////////////////////////////////////
    // Run whap
    if (par::debug) cerr << "** About to run whap... **\n\n";
    // run normally unless --largest is in effect
    if (!par::hap_specific)
	win[Window::n]->results = whap(*this,nhap);
    else
    {
	if (nhap < 2)
	{
	    WindowResults temp;
	    int p = par::perm+1;
	    temp.primary.resize(p);
	    for (int i=0; i<p; i++)
		temp.primary[i] = -1;
	    win[Window::n]->results = temp;
	}
	else
	{
	    // enable user-constraint mode
	    par::usr_constraint = true;
	    
	    WindowResults final;
	    vector<WindowResults> specific;
	    
	    // consider each specific haplotype
	    for (int i=1; i<=nhap; i++)
	    {
  	      
	        hs = i;
	        
                usr_alt.clear();
		usr_null.clear();
		
		usr_alt.resize(nhap);
		usr_null.resize(nhap);
		
		// first element must be 1
		usr_alt[0] = 1;
		usr_null[0] = 1;
		
		// set up appropriate constraint list
		for (int j=2; j<=nhap; j++)
		{
		    if (i==1) usr_alt[j-1] = 2;
		    else if (i==j) usr_alt[j-1] = 2;
		    else usr_alt[j-1] = 1;
		    usr_null[j-1] = 1;
		}
		
		// Analyse and store all results
		specific.push_back(whap(*this,nhap));
	    } 
	    
	    // generate final list of highest scoring haps from each permutation
	    // ignore secondary tests when constrained
	    
	    int p = specific[0].primary.size();
	    final.primary.resize(p);
	    
	    for (int i=0; i<p; i++)
	    {
		// get max for each perm
		final.primary[i] = specific[0].primary[i];
		for (int j=1; j<nhap; j++)
		{
		    if (specific[j].primary[i] > final.primary[i] )
			final.primary[i] = specific[j].primary[i];
		}
	    }
	    
	    win[Window::n]->results = final;
	}
    }
}
 



void displayPhase(CPhase & p)
 {

     p.paternal.display();
     cout << "\\";
     p.maternal.display();
     cout << "\t" << p.prob;     
     if (p.prob==0) cout << " * ";
     cout << "\t{";
     
     map<CHaplotype,int>::iterator patit = ind::hcode.find(p.paternal);
     map<CHaplotype,int>::iterator matit = ind::hcode.find(p.maternal); 
     
     if ( patit != ind::hcode.end() )
	 cout << ind::hapfreq1[patit->second];
     else
 	 cout << "------";
     
     cout << "\\";
     
     if ( matit != ind::hcode.end() )
 	 cout << ind::hapfreq1[matit->second];
     else
 	 cout << "------";
     
     cout << "}";

 }
