#include "function.h"
#include "haplotype.h"
#include "parameters.h"
#include "driver.h"
#include "trios.h"

CFunction::CFunction(CSample &s, Model &m, Driver &d)
    : sample(s), model(m), driver(d) 
{ 
} 

double CFunction::func(CVector &param)
{
    
    // order of parameters : mean/prev var covs bet with freqs 
    int cnt = 1;
    
    // Prevalence if binary, mean if continuous
    if (!par::fix_mean)
	constant = param[cnt++];
    
    //get variance if a continuous trait
    if (!par::binary && !par::fix_var) 
    {
	resvar = param[cnt] * param[cnt];
	cnt++;
    }
    
    // any covariates included?
    if (model.cov) 
    {
	cbeta.setCol(par::covn);
	for (int i=1; i<=par::covn; i++)
	    cbeta[i] = param[cnt++];
    }

    // any GxE moderators
    CMatrix i_tempbeta(par::covn,model.nsubhap);
    model.ibeta.setRow(par::covn);
    model.ibeta.setCol(model.nhap);
    if (model.gxe)
    {
	for (int c=1; c<=par::covn; c++)
	{
	    i_tempbeta[c][1] = 0; // first hap GxE effect is fixed
	    for (int h=2; h<=model.nsubhap; h++)
		i_tempbeta[c][h] = param[cnt++];
	}
    }
    // ...and expand
    for (int c=1; c<=par::covn; c++)
	for (int h=1; h<= model.nhap; h++)
	    model.ibeta[c][h] = i_tempbeta[c][model.betacode(h)];
    
    // Between effects
    bet_beta.setCol(model.nhap);
    CColVector bet_tempbeta(model.nsubhap);

    // Within effects 
    with_beta.setCol(model.nhap);
    CColVector with_tempbeta(model.nsubhap);

    // Extract reduced set of betas
    if (model.total)
    {
	bet_tempbeta[1] = with_tempbeta[1] = 0;
	for (int i = 2 ; i <= model.nsubhap ; i++) 
	    bet_tempbeta[i] = with_tempbeta[i] = param[cnt++];
    }
    else if (model.between)
    {
	bet_tempbeta[1] = 0;
	for (int i = 2 ; i <= model.nsubhap ; i++) 
	    bet_tempbeta[i] = param[cnt++];
    }
    
    if (model.within && !model.total)
    {
	with_tempbeta[1] = 0;
	for (int i = 2 ; i <= model.nsubhap ; i++) 
	    with_tempbeta[i] = param[cnt++];
    }

    // expand betas, imposing constraints as defined in betacode()
    for (int i = 1 ; i <= model.nhap ; i++) 
    {
	bet_beta[i] = bet_tempbeta[model.betacode(i)];
	with_beta[i] = with_tempbeta[model.betacode(i)];	
    }


    // Frequencies, freq_beta()
    freq_beta.setCol(model.nhap);
    
    if (par::joint)
    {
	// extract... using logisitc function to ensure 0->1 range
	for (int i = 1 ; i <= model.nhap ; i++) 
	    freq_beta[i] = 1/(1+exp(-param[cnt++]));

	// and ensure they sum to 1
	double fsum = freq_beta.sum();
	freq_beta.multiply(1/fsum);
			      			      
	// copy to indfreq1[]
	ind::hapfreq1 = freq_beta;
    }
    else
    {
	// ...or fix
	freq_beta = ind::hapfreq1;
    }

    // mean, var, cbeta, ibeta, bet_beta, with_beta, freq_beta;


   ////////////////////////////////////////////
   // Recalculate probabilities for everyman and everytrio 
   // if joint analysis is in effect
    
    if (par::joint)
    {
	// All possible individuals
	if (!par::no_unrel)
	    driver.scoreWithoutParents(driver.poss,
				       driver.everyman->box->K,
				       driver.everyman->box->B,
				       driver.everyman->box->W,
				       driver.everyman->box->nphase,
				       model.nhap,
				       driver.everyman,
				       0);

	// All possible trios
	if (!par::no_trios)
	    driver.scoreWithParents(driver.pposs,
				    driver.everytrio->box->K,
				    driver.everytrio->box->B,
				    driver.everytrio->box->W,
				    driver.everytrio->box->nphase,
				    model.nhap,
				    driver.everytrio,
				    0);
    }


    ///////////////////////////////////////////////
    // Qualitative conditional speedup
       
    if (par::qual_cond_speedup)
    {

	// nb. can speed up further by saying P(D) is fixed prevalence
	
	// Individuals without parents
	if (!par::no_unrel)
	{
	    // Unaffected individual
	    driver.everyman->bt(1) = false;
	    false_ind_denom = calculateLikelihood(driver.everyman);
	    // Affected individual
	    driver.everyman->bt(1) = true;
	    true_ind_denom = calculateLikelihood(driver.everyman);
	}

	// Individuals with parents
	if (!par::no_trios)
	{
	    // Unaffected trio offspring
	    driver.everytrio->bt(1) = false;
	    false_trio_denom = calculateLikelihood(driver.everytrio);
	    // Affected trio offspring
	    driver.everytrio->bt(1) = true;
	    true_trio_denom = calculateLikelihood(driver.everytrio);
	}
    }

    
    /////////////////////////////
    // Calculate Likelihood 
    
    long double log_likelihood = 0 ;
    int n = sample.countFamily();
    
    
    ////////////////////////////////////
    // MAIN LOOP : consider each proband

    for (int i = 1 ; i <= n ; i++) 
    {

	// get proband
	proband = sample.fam(i)->proband;

	// has non-missing sample data?
	if (!par::binary)
	    if (proband->qt(1) == CSample::missing_pheno) continue;
	if (par::binary)
	    if (proband->bt(1) == CSample::missing_pheno) continue;
	
	
	// if in --joint model, need to rescore K, B and W
	if (par::joint)
	    if (!proband->box->parents)
	    {
		
		// nb. when --joint is in effect, it may be better just to 
		//     rescore K, rather than the entire scoring
		driver.scoreWithoutParents(proband->box->P,
					   proband->box->K,
					   proband->box->B,
					   proband->box->W,
					   proband->box->nphase,
					   model.nhap,
					   proband,
					   par::uninform_threshold);
	    }
	    else
	    {
		driver.scoreWithParents(proband->box->PP,
					proband->box->K,
					proband->box->B,
					proband->box->W,
					proband->box->nphase,
					model.nhap,
					proband,
					par::uninform_threshold);
	    }
	
	

	/////////////////////////////////////////////
	// GxE INTERACTION
	
	if (model.gxe)
	{
	    // restore original between and within betas for each individual...
	    for (int i = 1 ; i <= model.nhap ; i++) 
	    {
		bet_beta[i] = bet_tempbeta[model.betacode(i)];
		with_beta[i] = with_tempbeta[model.betacode(i)];	
	    }
	    
	    // ...and then modify 
	    for ( int h = 1 ; h <= model.nhap ; h++)
		for ( int c = 1 ; c <= sample.countCovariate(); c++)
		{
		    if (model.between || model.total)
			bet_beta[h] += model.ibeta[c][h] * proband->cov(c);
		    if (model.within || model.total)
			with_beta[h] += model.ibeta[c][h] * proband->cov(c);
		}
	}
	
	
	/////////////////////////////////////////////
	// NUMERATOR
	
	long double ind_likelihood = calculateLikelihood(proband);	
	

	/////////////////////////////////////////////
	// DENOMINATOR -- under conditional approach
	//             -- for quantitative traits 
	//             -- qualitative traits can be calculated
	//                above when not joint, no covariates, etc
	//                in which case par::qual_cond_speedup == true
	
 	if (par::cond)
 	{	    

	    long double denom;

	    // Binary traits with no covariates: 
	    if (par::qual_cond_speedup) 
	    {
		
		if (!proband->box->parents) 
		{
		    if (proband->bt(1)) denom = true_ind_denom;
		    else denom = false_ind_denom;
		}
		else
		{
		    if (proband->bt(1)) denom = true_trio_denom;
		    else denom = false_trio_denom;
		}
	    }
	    // Use individual-specific trait value
	    // e.g qualitative traits, if covariates specified
	    else
	    {
		// copy trait value across from proband
		if (!proband->box->parents)
		{
		    if (!par::binary)
			driver.everyman->qt(1) = proband->qt(1);
		    else
			driver.everyman->bt(1) = proband->bt(1);		    
		}
		else
		{
		    if (!par::binary)
			driver.everytrio->qt(1) = proband->qt(1);
		    else
			driver.everytrio->bt(1) = proband->bt(1);		    
		}
		
		// copy any covariates
		if (model.cov)
		    for (int c=1; c<=driver.sample.countCovariate(); c++)
		    {
			if (!proband->box->parents)
			    driver.everyman->cov(c) = proband->cov(c);
			else
			    driver.everytrio->cov(c) = proband->cov(c);
		    }
		
		// calculate likelihood for everyman / everytrio
		if (!proband->box->parents) denom = calculateLikelihood(driver.everyman);
		else denom = calculateLikelihood(driver.everytrio);
	    }		
	

	    // nb. need to change this to avoid FP exceptions 
	    if (denom == 0) 
		ind_likelihood = 0;
	    else 
		ind_likelihood /= denom;
	    
	} // end of DENOMINATOR calculation
	
	
	// Accumulate sample log likelihood
	if (ind_likelihood>0) 
	    log_likelihood += -2 * log(ind_likelihood);
	
    } // end if INDIVIDUAL LOOP
    
    return (double)log_likelihood;
    
}

    
/////////////////////////////////////////////////////
// Helper functions for calculation of likelihood

double CFunction::calculateExpected()
{
    
    double expected = 0;
    
    // covariates
    // nb. for multiple covariates, strictly speaking we also need
    //     to consider the covariance between them . 
    if (model.cov)
	for ( int c = 1 ; c <= sample.countCovariate(); c++)
	    expected += cbeta[c] * sample.cov(c)->mean;
    
    // calculate between effects
    for ( int h = 1 ; h <= model.nhap ; h++)
	expected += 2 * bet_beta[h] * freq_beta[h];

    return expected;

}

double CFunction::calculateVariance()
{
    
    double variance = 0;
    
    // covariates
    if (model.cov)
	for ( int c = 1 ; c <= sample.countCovariate(); c++)
	    variance += cbeta[c]*cbeta[c] * sample.cov(c)->var;
    
    // between effects
    double bmean = 0;
    for ( int h = 1 ; h <= model.nhap ; h++)
	bmean += bet_beta[h] * freq_beta[h];

    for ( int h = 1 ; h <= model.nhap ; h++)
    {
	variance += 2*mymath::sqr( bet_beta[h] - bmean) * freq_beta[h];
    }

    // make sure that the calculate variance doesn't exceed the total 
    if (variance > par::fixed_var) return par::fixed_var-(par::fixed_var*0.05);
    else return variance;
        
}


/////////////////////////////////////////////////////////////////////
// Calculate weighted L(X|G)

long double CFunction::calculateLikelihood(CIndividual * person)
{
    
    long double ind_likelihood = 0 ;
    
    // number of phases
    nphase = person->box->K.getCol();
    
    // consider each phase
    for (int j = 1 ; j <= nphase; j++)
    {

	double m = 0 ;
	
	// covariate effects
	if (model.cov) 
	    for ( int k = 1 ; k <= sample.countCovariate() ; k++)
		m += cbeta[k] * person->cov(k);
	
	// haplotype effects
	// either consider all B and W effects for all individuals
	// or constrain B component in trios to 0 
	if (!par::no_between_trios || !person->box->parents )
	{
	    for ( int k = 1 ; k <= model.nhap ; k++)
		m += bet_beta[k] * person->box->B[j][k]
		    + with_beta[k] * person->box->W[j][k]; 
	}
	else if (par::no_between_trios && person->box->parents)
	{
	    for ( int k = 1 ; k <= model.nhap ; k++)
		m += with_beta[k] * person->box->W[j][k];	    
	}

	// Fix mean?
	if (!par::fix_mean) 
	{ 
            // in this case, constant defined above as param[1]
	    m += constant;
	}
	else
	{
	    //...otherwise, we have to calculate the constant term 
	    // calculate expected value: (c meanC + meanB)  (as meanW=0)
	    constant = par::fixed_mean - calculateExpected();
	    m += constant;
	}

	// Fix (residual) variance
	if(par::fix_var)
	    resvar = par::fixed_var - calculateVariance();

	// calculate likelihood...
	// ...for BINARY traits
	if (par::binary)
	{
	    long double m2 = 1 / ( 1 + exp(-m) );
	    
	    if (person->bt(1))
		ind_likelihood += m2 * person->box->K[j];
	    else
		ind_likelihood += (1 - m2) * person->box->K[j];
	}
	// ...or for QUANTITATIVE traits
	else 
	{
	    ind_likelihood += mymath::normden(person->qt(1) , m , resvar) * person->box->K[j];
	}
	
    } // end of CONSISTENT PHASE LOOP
    
    return ind_likelihood;
    
}

