#include "reader.h"
#include "parameters.h"
#include "trios.h"
#include <cstdio>

// Extract haplotype information from output of SNPHAP

// snphap data.raw file1.out file2.out -th 0 -ss
// -th 0   Threshold = 0 (i.e. all possible haplotypes) 
// -ss     Spreadsheet-style output

// Generate K to be approproiate size for each individual 
// (i.e. nphase is individual specific)

// Format : ID chr  Alleles P()
// Number of SNPs - #cols - 3 therefore (ID, P(), chr)


int runsnphap(CSample & sample,
	      IntArray &id1,
	      StringArray &id2,
	      StringArray &id3,
	      Driver &drv,
	      string out_name)
{
    
// run SNPHAP (David Clayton)
    
    int intth = int(par::threshold * 100);
    if (intth > 100) intth = 100;
    if (intth < 0 ) intth = 0;
    
    string command;
    if (intth == 100) 
	command = "snphap "+out_name+" "+out_name+".o1 "+out_name+".o2 -ss -th 1 > " +out_name+".dummy";
    else
	command = "snphap "+out_name+" "+out_name+".o1 "+out_name+".o2 -ss -th 0." 
	    + int2str(intth) + " > " +out_name+".dummy";
    
    bool failed = system(command.c_str());
    if (failed) error("SNPHAP did not run properly... check your data");

    string tmp = out_name+".dummy";
    remove(tmp.c_str());
    
    if (par::debug) cerr << "\nsnphap() finished\n";


    ////////////////////////////////////////////////////
    // Read in haplotype list, frequencies
    
    CData freqdat;
    freqdat.header= true;
    freqdat.import(out_name+".o1");
    tmp = out_name+".o1";
    if (!par::keep_snphap_files)
	remove(tmp.c_str());
    
    ////////////////////////////////////////////////////
    // Read in posterior probability list
    
    CData dat;
    dat.header = true;
    dat.import(out_name+".o2");
    tmp = out_name+".o2";
    if (!par::keep_snphap_files)
	remove(tmp.c_str());

    int nSNP = dat.getVar() - 3;
    ind::nsnps = nSNP;
    if (!par::usat) ind::nloci = nSNP;
    else ind::nloci = drv.win[Window::n]->alt.size();
 
   
    ///////////////////////////////////////////////////
    // Make a list of common haplotypes

    // these are always sorted by frequency in SNPHAP, so 
    // we can stop when we reach 

    vector<CHaplotype> haplotype_list;
    haplotype_list.clear();
    
    for (int i=1; i<=freqdat.getInd(); i++)
    {

	// is the haplotype common enough?
	if (freqdat[i][freqdat.getVar()] >= par::abs_threshold) 
	{
	    haplotype_list.push_back(CHaplotype(nSNP));
	    for (int k = 1 ; k <= nSNP ; k++)
		haplotype_list[haplotype_list.size()-1].pos(k) = int2str((int)freqdat[i][k]);
	}
    }
    
    // remove any haplotypes that do not conform with usat downcoding
    
    //   ac[indx]  0  1  2      3  4        5
    //   locus;    1  2  3      4  5        6
    //   acount;   1  1  3      1  4        1
    //             1  1  1 2 1  2  1 1 2 2  1
    //                             XXXXXXX
    //                             e.g. sums to 6 not 5

    if (par::usat)
    {
	for (int i=0; i< haplotype_list.size(); i++)
	{
	    bool problem = false;
	    int n=0;

	    for (int k=0; k < drv.win[Window::n]->alt.size(); k++)    
	    {
		int a = drv.win[Window::n]->alt[k]-1;
		    
		// if it is a SNP, we can move on
		    if (drv.acount[a]==1) n++;
		    else // otherwise we must check sanity
		    {
			// if coded properly, should sum to k+1
			int s=0;
			for (int k2=0; k2<drv.acount[a]; k2++)
			    s += atoi(haplotype_list[i].pos(++n).c_str());
			if ( s != drv.acount[a]+1 ) problem = true;
		    }
	    }
	    
	    // if inconsistent...
	    if (problem) 
		{
		    // ... remove
		    haplotype_list.erase(haplotype_list.begin()+i);
		    // ... and repeat this index
		    i--;
		}
	}
    }
    
    int nhap = haplotype_list.size();
    if (par::debug) cerr << "\n" << nhap << " haplotypes found (of " << freqdat.getInd() << ")\n";
    
    
    ///////////////////////////////////////////////////////////////
    // Run through all individual posterior probabilities

    int row = 0;
    int indiv = 0; 
    int current_id = (int)dat[1][1];
    bool finished = false;

    while (1) 
    {
	vector<int> indiv_hap;
	vector<double> indiv_prob;
	indiv_hap.clear();
	indiv_prob.clear();
	int indiv_nphase = 1;
	
	if (par::debug)
	{
	    cerr << "\n processing " << current_id << "\n";
	    cerr << "\n row =  " << row << "\n";
	}

	// for this founder
	while ((int)dat[++row][1] == current_id)
	{

	    CIndividual * person = sample.getIndividual( id2(current_id), 
							 id3(current_id));

	    // which haplotype?
	    bool none_match = true;
	    for (int i=0 ; i<nhap; i++)
	    {
		bool match = true;
		for (int j=1 ; j <= nSNP ; j++)
		    if ( haplotype_list[i].pos(j) != int2str((int)dat[row][j+2]) ) match = false;
		// found the match
		if (match) 
		{
		    if (par::debug)
			cerr << "adding i=" << i << "\n";
		    
		    none_match = false;
		    indiv_hap.push_back(i);
		    indiv_prob.push_back(dat[row][nSNP+3]);
		    break;
		}
	    }   
	    if (none_match) // haplotype been exlcuded (too rare)
	    { 
		indiv_hap.push_back(-1);
		indiv_prob.push_back(dat[row][nSNP+3]);
	    }

		
            // end of file? 
	    if (row >= dat.getInd() ) 
	    { finished = true; break; }
	}

	if (par::debug)
	{
	    cerr << "\n BEFORE : going to remove some phases\n";
	    for (int i=0; i< indiv_hap.size() ; i++)
		cerr << "\n i=" << i << " indiv_hap[i] = " << indiv_hap[i] << "\n";
	}

	
	// remove phases containing at least one rare haplotye
        // and recalculate phase probabilities so they sum to 1

	// copy of valid phases
	// haplotypes should be in phase pairs
	vector<int> indiv_hap2;
	vector<double> indiv_prob2;
	for (int i=0; i< indiv_hap.size() ; i+=2)
	    if (indiv_hap[i] != -1 && indiv_hap[i+1] != -1)
	    {
		// add only non-zero phases 
		if (indiv_prob[i] > 0)
		{
		    indiv_hap2.push_back(indiv_hap[i]);
		    indiv_hap2.push_back(indiv_hap[i+1]);
		    indiv_prob2.push_back(indiv_prob[i]); // only one copy of this

		    if (par::debug)
		    {
			cerr << "\n adding indiv_hap[i], indiv_hap[i+1], indiv_prob[i] = "  
			     << indiv_hap[i] << "   " 
			     << indiv_hap[i+1] << "   " 
			     << indiv_prob[i] << "\n";		
		    }
		}
	    }
	
	indiv_nphase = indiv_hap2.size() / 2; 
	

	if (par::debug)
	{
	    cout << "\n indiv_nphase = " << indiv_nphase << "\n";
	    cout << "\n indiv_prob2.size() = " << indiv_prob2.size() << "\n";
	    cout << "\n indiv_hap2.size() = " << indiv_hap2.size() << "\n";
	    cout << "\n haplotype_list.size() = " << haplotype_list.size() << "\n";
	}

	// Identify individual
	CIndividual * person = sample.getIndividual(id2(current_id),id3(current_id));
	
	// Adjust for excluded rare haps: make frequencies sum to 1
	// (remove phases that have zero probability below -- this will not 
        //  effect the summation here)
	// Note -- because of limited decimal places in the whapout2 file, 
	//         we can expect exactly zero, which is fine.
	double indiv_prob_sum = 0;
	for (int i=0;i<indiv_prob2.size();i++) indiv_prob_sum += indiv_prob2[i];
	
	// remove any existing phases
	person->box->clearPhase();

	// add Phases (split into the two possible parental/maternal phases
	// with equal probability)
	for (int i=0; i<indiv_hap2.size() ; i+=2 )
	    if (indiv_prob_sum > 0)
	    {

		if (par::debug)
		{
		    cerr << "\n entering this phase ..." << i << "  \n"; 
		    cerr << "\nindiv_hap2[i]] = " << indiv_hap2[i];
		    cerr << "\nindiv_hap2[i+1] = " << indiv_hap2[i+1];
		    cerr << "\n";
		}

		// if Homozygote add once
		if ( haplotype_list[indiv_hap2[i]] == haplotype_list[indiv_hap2[i+1]] ) 
		{
		    if (par::debug)
			cerr << "\n as a homo : i = " << i << " and int(i/2)= " << int(i/2) << "\n";

 		    person->box->addPhase(haplotype_list[indiv_hap2[i]],
					  haplotype_list[indiv_hap2[i+1]],
					  (indiv_prob2[int(i/2)]/indiv_prob_sum));
		}
		else
		// if Heterozygote add twice
		{
		    if (par::debug)
			cerr << "\n as a het : i = " << i << " and int(i/2)= " << int(i/2) << "\n";

 		    person->box->addPhase(haplotype_list[indiv_hap2[i]],
					  haplotype_list[indiv_hap2[i+1]],
					  (indiv_prob2[int(i/2)]/indiv_prob_sum) / 2);
		    
		    person->box->addPhase(haplotype_list[indiv_hap2[i+1]],
					  haplotype_list[indiv_hap2[i]],
					  (indiv_prob2[int(i/2)]/indiv_prob_sum) / 2);
		}
		
	    }
	
	// update ID of new individual
	current_id = (int)dat[row][1];
	// and subtract row count
	row--;
	indiv++;
	
	// loop to next individual, unless finished
	if (finished) break;
    }
    
    if (par::debug) cerr << "\nfinished reading in posterior probs\n";


    // X-chromosome capability still under construction...
    // NOTE1 : this needs to be changed so that previously unobserved haps can be assigned here
    // NOTE2 : this needs to be changed for uSat compatibility
    // make perfect phased haplotypes for male X chromosome individuals
    if (par::xchr)
    {
	for (int f=1; f<=sample.countFamily();f++)
	    for (int i=1; i<=sample.fam(f)->nInFamily();i++)
	    {
		CIndividual * person = sample.fam(f)->ind(i);
		if (person->sex == 1)
		{
		    
		    // remove any existing phases
		    person->box->clearPhase();

		    cout << "nSNP = " << nSNP << "\n";
		    cout << "window size = " << drv.win[Window::n]->alt.size() << "\n";

		    // Create haplotype
		    CHaplotype x(nSNP);
		    for (int k=1; k <= drv.win[Window::n]->alt.size(); k++)    
		    {
			x.pos(k) = person->loc(drv.win[Window::n]->alt[k-1]).one;
			cerr << x.pos(k) << ": win[]->alt[] = " << drv.win[Window::n]->alt[k-1] 
			     << person->loc(drv.win[Window::n]->alt[k-1]).one << "\n";
		    }

		    // Assign, as if diploid, perfect probability
 		    if (par::debug) cerr << "\n added phase for male X chr \n";
		    person->box->addPhase(x,x,1);
		    
		}
	    }
	
    }


    // NOTE3 : the routine(s) below need to be adapted for Xchr...
    // i.e. so that trio data can be analysed with Xchr...

    // phase kids in trios given parents
    phaseChildGivenParents(drv);

    if (par::debug) cerr << "\nfinished phasing kids given parents \n";
        
    // Calculate Sample haplotype names 
    ind::hap.resize(nhap+1);  // zero-based but ignore first element... 
    for (int i=1 ; i<=nhap; i++)
    {
	ind::hap[i].length(nSNP);
	for (int j=1 ; j <= nSNP ; j++)
	    ind::hap[i].pos(j) = haplotype_list[i-1].pos(j);	
    }
    
    
    // Extract sample (pre-processing) frequencies from whapout1
    // and set codes
    ind::hapfreq.setCol(nhap);    
    ind::hapfreq1.setCol(nhap);
    ind::hapfreq.clear();
    ind::hapfreq1.clear();
    ind::hcode.clear();

    // for each haplotype
    for (int i=0 ; i < nhap ; i++)
    {
	// find position in whapout1 (freqdat)
	
	for (int j = 1 ; j <= freqdat.getInd() ; j++)
	{
	    bool match = true;
	    for (int k = 1 ; k <= nSNP ; k++)
	    {
		if (haplotype_list[i].pos(k) != int2str((int)freqdat[j][k])) 
		    match = false;
	    }
	    if (match) 
	    {
		ind::hapfreq[i+1] = freqdat[j][freqdat.getVar()];
		ind::hcode.insert(make_pair(haplotype_list[i],i+1));
		break;
	    }
	}
    }

    // create ind::hapfreq1 that sums to 1
    double s = ind::hapfreq.sum();
    ind::hapfreq1 = ind::hapfreq;
    ind::hapfreq1.multiply(1/s);
      
//    if (nhap <= 1) error("Need >=2 haplotypes w/ freq > --at value in sample");


    return nhap;

}

