#include <iomanip>

#include "results.h"
#include "haplotype.h"
#include "parameters.h"
#include "helper.h"


double results_calculateExpected(Driver &,Model &);
double results_calculateVariance(Driver &,Model &);

/////////////////////////////////////////////////////////////////////
// Display results for a specific window in detail
//
// Otherwise, summary sliding window results are handled by Driver


void disp_results(Driver & d, Model & alt, Model & null, 
		  vector<double> & permutation_lrt,
		  vector< vector<double> > & permutation_secondary)
{
    
    ////////////////////////////////////////
    // Calculate frequency of subhaplotypes
    CVector subhapfreq(alt.nsubhap);
    helper::subFrequency(alt,subhapfreq);

    ////////////////////////////////////////////////////////
    // Make subhaplotype -> haplotype conversion table 
    IntArray sub2full;
    helper::sub2fullMapping(alt,sub2full);

    /////////////////////////////////////
    // Get betacodes for null model 
    IntArray null_betacode2;
    null_betacode2().resize(alt.nsubhap);
    for (int i = 1 ; i <= alt.nsubhap ; i++) 
	for (int j = 1 ; j <= alt.nhap ; j++) 
	    if (alt.betacode(j) == i)
	    {	
		null_betacode2(i) = null.betacode(j);
		break;
	    }
    

    /////////////////////////////////////
    // DISPLAY RESULTS
    /////////////////////////////////////
    
    cout << setiosflags(ios::left) << "\n";
    cout << setw(ind::nloci+6) << "Hap"
	 << setw(8) << "Freq" 
	 << setw(8) << "Alt(B)" 
	 << setw(8) << " "
	 << setw(8) << "Alt(W)" 
	 << setw(8) << " "
	 << setw(8) << "Null(B)" 
	 << setw(8) << " "
	 << setw(8) << "Null(W)" 
	 << "\n";
    
    cout << setw(ind::nloci+6) << "---"
	 << setw(8) << "-----" 
	 << setw(8) << "------" 
	 << setw(8) << " "
	 << setw(8) << "------" 
	 << setw(8) << " "
	 << setw(8) << "-------" 
	 << setw(8) << " "
	 << setw(8) << "-------" 
	 << "\n";
    
    
    // VERBOSE OUTPUT -- LIST ALL FULL HAPLOTYPES
    // NECESSARY WHEN A USER --CONSTRAINT IS IN EFFECT
    if (par::verbose || par::usr_constraint)
    {
	string alt_flag = "["+int2str(alt.betacode(1))+"]";
	string null_flag = "["+int2str(null.betacode(1))+"]";

	string hapname = helper::getHapName(d,ind::hap[1]);

	cout << setw(ind::nloci+6) << hapname;
	cout << setw(8) << ind::hapfreq1[1] 
	     << setw(8) << "0.000" 
	     << setw(8) << " "
	     << setw(8) << "0.000" 
	     << setw(8) << alt_flag
	     << setw(8) << "0.000" 
	     << setw(8) << " "
	     << setw(8) << "0.000" 
	     << setw(8) << null_flag
	     << "\n";
	
	for (int i = 2 ; i <= alt.nhap ; i++) 
	{

	    string alt_flag = "["+int2str(alt.betacode(i))+"]";
	    string null_flag = "["+int2str(null.betacode(i))+"]";
	    string hapname = helper::getHapName(d,ind::hap[i]);
	
	    cout << setw(ind::nloci+6) << hapname;
	    cout << setw(8)            << ind::hapfreq1[i]; 
	    cout << setw(8) << alt.bet_beta[i]
		 << setw(8) << " "
		 << setw(8) << alt.with_beta[i]
		 << setw(8) << alt_flag
		 << setw(8) << null.bet_beta[i]
		 << setw(8) << " " 
		 << setw(8) << null.with_beta[i]
		 << setw(8) << null_flag;
	    cout << "\n";
	    
	}
    }
	
    // NONVERBOSE OUTPUT -- LIST ALL ONLY SUBHAPLOTYPES
    if (!par::verbose && !par::usr_constraint)
    {
	
	// CHECK: should not use --all and --usat options together
	//        otherwise output may be corrupted......  (?)
	
	// generate subhaplotype string
	string subhapname = helper::getHapName(d,ind::hap[sub2full(1)]);

	string alt_flag = "["+int2str(alt.betacode(1))+"]";
	string null_flag = "["+int2str(null.betacode(1))+"]";
	
	cout << setw(ind::nloci+6) << subhapname
	     << setw(8) << subhapfreq[1] 
	     << setw(8) << "0.000"
	     << setw(8) << " "
	     << setw(8) << "0.000" 
	     << setw(8) << alt_flag
	     << setw(8) << "0.000" 
	     << setw(8) << " "
	     << setw(8) << "0.000" 
	     << setw(8) << null_flag
	     << "\n";

	for (int i = 2 ; i <= alt.nsubhap ; i++) 
	{
	    
	    // generate subhaplotype string
	    string subhapname = helper::getHapName(d,ind::hap[sub2full(i)]);
	    string alt_flag = "["+int2str(alt.betacode(sub2full(i)))+"]";
	    string null_flag = "["+int2str(null.betacode(sub2full(i)))+"]";

	    cout << setw(ind::nloci+6) << subhapname 
		 << setw(8)            << subhapfreq[i];
	    
	    cout << setw(8) << alt.bet_beta[sub2full(i)]
		 << setw(8) << " "
		 << setw(8) << alt.with_beta[sub2full(i)]
		 << setw(8) << alt_flag
		 << setw(8) << null.bet_beta[sub2full(i)]
		 << setw(8) << " " 
		 << setw(8) << null.with_beta[sub2full(i)]
		 << setw(8) << null_flag;
	    cout << "\n";
	    
	}
	
    }
    

    cout << setw(ind::nloci+6) << "---"
	 << setw(8) << "-----" 
	 << setw(8) << " " 
	 << setw(8) << " "
	 << setw(8) << "------" 
	 << setw(8) << " "
	 << setw(8) << " "
	 << setw(8) << " "
	 << setw(8) << "-------" 
	 << "\n";

    cout << setw(ind::nloci+6) << " "
	 << setw(8) << " " 
	 << setw(8) << " " 
	 << setw(8) << " "
	 << setw(8) << alt.likelihood
	 << setw(8) << " "
	 << setw(8) << " " 
	 << setw(8) << " "
	 << setw(8) << null.likelihood 
	 << "\n";
    
    
    ////////////////////////////////////////////
    // Covariates
    
    if (alt.cov)
    {
	cout << "\n";    
	
	for (int i=1; i <= par::covn; i++)
	{
	    cout << setw(ind::nloci+6) << "cov " << i 
		 << setw(8) << " " 
		 << setw(8) << " " 
		 << setw(8) << " "
		 << setw(8) << alt.cbeta[i];
	    if (null.cov)
		cout << setw(8) << " "
		     << setw(8) << " " 
		     << setw(8) << " "
		     << setw(8) << null.cbeta[i];
	    cout << "\n";    
	}
    }


    ////////////////////////////////////////////
    // GxE moderators
    
    if (alt.gxe)
    {
	cout << "\n";    
	
	for (int c=1; c <= par::covn; c++)
	{

	    for (int h=1; h <= alt.nhap; h++)
	    {
		cout << setw(ind::nloci+6) << "gxe: cov " << c << " x hap " << h 
		     << setw(8) << " " 
		     << setw(8) << " "
		     << setw(8) << alt.ibeta[c][h];
		if (null.gxe)
		    cout << setw(8) << " "
			 << setw(8) << " " 
			 << setw(8) << " "
			 << setw(8) << null.ibeta[c][h];
		cout << "\n";    
	    }
	}
    }

    
    ////////////////////////////////////////////
    // Mean, prevalence, variance
    
//     if (!par::fix_mean)
// 	cout << "\nConstant         = " << alt.mean << "\t" << null.mean;
//     else
// 	cout << "\nConstant         = " << par::fixed_mean -  results_calculateExpected(d,alt)
// 	     << "\t" << par::fixed_mean - results_calculateExpected(d,null);
    
//     if (!par::binary)
//     {
// 	if (!par::fix_var)	
// 	    cout << "\nLocus % variance = " 
// 		 << results_calculateVariance(d,alt)/(results_calculateVariance(d,alt)+alt.var) 
// 		 << "\t" 
// 		 << results_calculateVariance(d,null)/(results_calculateVariance(d,null)+null.var);
// 	else
// 	    cout << "\nLocus % variance = " 
// 		 << results_calculateVariance(d,alt)/par::fixed_var
// 		 << "\t" 
// 		 << results_calculateVariance(d,null)/par::fixed_var;
//     }
    
//     cout << "\n";
    

    ////////////////////////////////////////////
    // Some information about coverage 

    cout << "\nProportion of haplotypes covered = " << ind::hapfreq.sum() << "\n"; 
    
    
    ////////////////////////////////////////////
    // Final fit statistics: LR and Wald tests
    
    double chisq = null.likelihood - alt.likelihood;
    int df = alt.nparam - null.nparam;
    
    if (df==0) chisq=0;
    if (chisq<0) chisq=0;
    
    cout << "LRT = " << chisq
	 << "\ndf  = " << df;
    
    if (!(par::one_sided_pos || par::one_sided_neg))
      {
	// two-sided test
	cout << "\np   = "
	     << setiosflags(ios::scientific)    
	     << mymath::nchi(chisq,df,0)
	     << resetiosflags(ios::scientific)
	     << "\n";
      }
    else
      {
	cout << "\np(1) = " 
	     << setiosflags(ios::scientific)    
	     << mymath::norm(sqrt(chisq),0,1)
	     << resetiosflags(ios::scientific)
	     << "\n";
	if (df>1) 
	  cout << "warning:: one-sided p-value given for omnibus test\n";
      }
    
    
    
    
    /////////////////
    // Confidence Intervals on parameters
    
    if (par::ci)
      {
	for (int i=1; i <= alt.estimates.getCol(); i++)
	  {
	    cout << "\n["<<i<<"] = " << alt.estimates[i] << "\t( " ;
	    double c = mymath::invnorm((1-par::civ)/2,0,1);
	    double lower = alt.estimates[i] - c * sqrt(alt.covbeta[i][i]);
	    double upper = alt.estimates[i] + c * sqrt(alt.covbeta[i][i]);
	    cout << lower << ",\t" << upper << " )";
	}
	cout << "\n";
    }


    /////////////////
    // Wald tests
    
    // NOTE: this whole section of code needs updating
    // NOTE: Wald tests not particularly useful in any case

//     if (par::wald)
//     {
		
// 	CColVector a(alt.estimates.getCol());
// 	for (int i=1; i<= a.getCol(); i++) a[i] = alt.estimates[i];
// 	CColVector h = a;
	
// 	if (par::binary) // in binary case, keep just mean
// 	    for (int i=2; i<= alt.estimates.getCol(); i++)
// 		h[i] = 0;
// 	else // in continuous case, keep mean and variance
// 	    for (int i=3; i<= alt.estimates.getCol(); i++)
// 		h[i] = 0;
	
// 	double W = (a-h).transpose() * alt.covbeta.inverse() * (a-h);
// 	cout << "\n\nFull Wald test = " << W << "\n";
	
// 	for (int l=1; l<=alt.estimates.getCol(); l++)
// 	{
// 	    //reset
// 	    for (int i=1; i<= a.getCol(); i++) a[i] = alt.estimates[i];
// 	    h = a;
	    
// 	    h[l] = 0;
	    
// 	    double W2 = (a-h).transpose() * alt.covbeta.inverse() * (a-h);
// 	    cout << "Wald test (param " << l << ") = " << W2 << "\n";
	    
// 	}
//     }


    ///////////////////////////////
    // Permutation test results

    if (!par::window || (par::window && par::win_all_results))
    {
	
	// mega-verbose output
	if (!par::window && par::win_all_results)
	{
	    cout << "ORIG:\t" << permutation_lrt[0];
	    if (par::secondary && alt.nsubhap > 2 ) cout << "\t" << permutation_secondary[0][0];
	    cout << "\n";
	}
	
	//
	// Permutation tests : Primary
	//

	if (par::perm > 0)
	{

	    // mega-verbose output: primary and secondary(global)
	    if (par::win_all_results)
	    {
		for (int i=1 ; i < permutation_lrt.size() ; i++)
		{
		    cout << "PERM:\t" << permutation_lrt[i];
		    if (par::secondary && alt.nsubhap > 2 ) cout << "\t" << permutation_secondary[i][0];
		    cout << "\n";
		}
		cout << "\n";
	    }
	    
	    int exceed = 0;
	    
	    // start from 1 (ie. i==0 is the actual result)
	    for (int i=1 ; i < permutation_lrt.size() ; i++)
		if ( permutation_lrt[i] > permutation_lrt[0] ) exceed++;
	    
	    double emp_primary_p = (double)(exceed+1)/((permutation_lrt.size()-1)+1) ;	
	    
	    cout << "\nPrimary analysis : empirical p-value from " 
		 << permutation_lrt.size()-1 
		 << " permutations = " ;
	    cout.precision(10);
	    cout << emp_primary_p;
	    cout.precision(3);
	    cout << "\n";
	    
	    //
	    // Permutation tests : Secondary
	    //

	    if (par::secondary && alt.nsubhap > 2 )
	    {
		
		int end_pos = permutation_secondary[0].size() ;
		
		// Full secondary perm results
		if (par::secondary_verbose)
		{
		    for (int pos=0; pos < end_pos; pos++)
		    {
			// start from 1 (ie. i==0 is the actual result)
			for (int i=1 ; i < permutation_secondary.size() ; i++)
			    cout << permutation_secondary[i][pos] << "\t";
			cout << "\n";
		    }
		}
		

		// Consider the different flavours of secondary test
		for (int pos=0; pos < end_pos; pos++)
		{
		    int exceed = 0;
		    
		    // start from 1 (ie. i==0 is the actual result)
		    for (int i=1 ; i < permutation_secondary.size() ; i++)
			if ( permutation_secondary[i][pos] > permutation_secondary[0][pos] ) exceed++;

		    // use (r+1)/(n+1) formula for empirical p-values 
		    double emp_secondary_p1 = (double)(exceed+1)/((permutation_secondary.size()-1)+1);
		    
		    vector<double> sectmp(permutation_secondary.size());
		    for (int i=0; i< permutation_secondary.size() ; i++)
			sectmp[i] = permutation_secondary[i][pos];
		    
		    double emp_secondary_p2 = helper::combinedEmpiricalPValue(permutation_lrt, sectmp);
		    
		    if (pos == 0 )      cout << "\nSecondary analysis: global   ";
		    else if (pos == 1 ) cout << "\nSecondary analysis: large-loc";
		    else                cout << "\nSecondary analysis: loc   " << pos-1 ;
		    
		    cout << " : Empirical p-value (" 
			 << permutation_secondary.size()-1 
			 << " permutations) = ";
		    cout.precision(10);
		    cout << emp_secondary_p1
			 << "  combined = " 
			 << emp_secondary_p2;
		    cout.precision(3);
		    cout << "  (S = " <<  permutation_secondary[0][pos] 
			 << " )";
		    
		}
		cout << "\n";
	    }
	}
    }    
}




void disp_hs_results(Driver & d, Model & alt, Model & null, 
		     vector<double> & permutation_lrt,
		     vector< vector<double> > & permutation_secondary)
{


  ////////////////////////////////////////
  // Calculate frequency of subhaplotypes
  CVector subhapfreq(alt.nsubhap);
  helper::subFrequency(alt,subhapfreq);
  
  ////////////////////////////////////////////////////////
  // Make subhaplotype -> haplotype conversion table 
  IntArray sub2full;
  helper::sub2fullMapping(alt,sub2full);
  
  /////////////////////////////////////
  // Get betacodes for null model 
  IntArray null_betacode2;
  null_betacode2().resize(alt.nsubhap);
  for (int i = 1 ; i <= alt.nsubhap ; i++) 
    for (int j = 1 ; j <= alt.nhap ; j++) 
      if (alt.betacode(j) == i)
	{	
	  null_betacode2(i) = null.betacode(j);
	  break;
	}
  
  
  /////////////////////////////////////
  // DISPLAY RESULTS
  /////////////////////////////////////
  
  if (d.hs == 1) 
    cout << "\n\n";
  
  cout << d.hs << "\t";
  string hapname = helper::getHapName(d,ind::hap[d.hs]);
  
  if (d.hs==1)
    {
      cout << hapname << "  ";
      cout << ind::hapfreq1[d.hs] << "  ";
      cout << -alt.bet_beta[2]
	   << " "
	   << -alt.with_beta[2] << "\t";
    }
  else
    {
      cout << hapname << "  ";
      cout << ind::hapfreq1[d.hs] << "  "; 
      cout << alt.bet_beta[d.hs]
	   << " "
	   << alt.with_beta[d.hs] << "\t";
    }
  
  // USING THE OTHER METHOD....
// 	    string subhapname = helper::getHapName(d,ind::hap[sub2full(i)]);
// 	    string alt_flag = "["+int2str(alt.betacode(sub2full(i)))+"]";
// 	    string null_flag = "["+int2str(null.betacode(sub2full(i)))+"]";

// 	    cout << setw(ind::nloci+6) << subhapname 
// 		 << setw(8)            << subhapfreq[i];
	    
// 	    cout << setw(8) << alt.bet_beta[sub2full(i)]
// 		 << setw(8) << " "
// 		 << setw(8) << alt.with_beta[sub2full(i)]
// 		 << setw(8) << alt_flag
// 		 << setw(8) << null.bet_beta[sub2full(i)]
// 		 << setw(8) << " " 
// 		 << setw(8) << null.with_beta[sub2full(i)]
// 		 << setw(8) << null_flag;
// 	    cout << "\n";
	    
	
    
    double chisq = null.likelihood - alt.likelihood;
    int df = alt.nparam - null.nparam;
    
    if (df==0) chisq=0;
    if (chisq<0) chisq=0;
    
    cout << "\t" << chisq;
      
    if (!(par::one_sided_pos || par::one_sided_neg))
      {
	// two-sided test
	cout << "\t"
	     << setiosflags(ios::scientific)    
	     << mymath::nchi(chisq,df,0)
	     << resetiosflags(ios::scientific);
      }
    else
      {
	cout << "\t" 
	     << setiosflags(ios::scientific)    
	     << mymath::norm(sqrt(chisq),0,1)
	     << resetiosflags(ios::scientific);
      }
    
    cout << "\n";
    
} // end of HS results






double results_calculateExpected(Driver &d, Model & m)
{
    
    double expected = 0;
    
    // covariates
    // NOTE: for multiple covariates, strictly speaking we also need
    //       to consider the covariance betwee them ... 
    if (m.cov)
	for ( int c = 1 ; c <= d.sample.countCovariate(); c++)
	    expected += m.cbeta[c] * d.sample.cov(c)->mean;
    
    // calculate between effects
    for ( int h = 1 ; h <= m.nhap ; h++)
	expected += 2 * m.bet_beta[h] * ind::hapfreq1[h];

    return expected;

}

double results_calculateVariance(Driver &d, Model & m)
{
    
    double variance = 0;
    
    // covariates
    if (m.cov)
	for ( int c = 1 ; c <= d.sample.countCovariate(); c++)
	    variance += m.cbeta[c]*m.cbeta[c] * d.sample.cov(c)->var;
    
    // between effects
    double bmean = 0;
    for ( int h = 1 ; h <= m.nhap ; h++)
	bmean += m.bet_beta[h] * ind::hapfreq1[h];

    for ( int h = 1 ; h <= m.nhap ; h++)
	variance += 2 * mymath::sqr( m.bet_beta[h] - bmean) * ind::hapfreq1[h];
    
    return variance;
        
}
