#include "driver.h"
#include "parameters.h"
#include "trios.h"
#include "helper.h"

// temp function
void displayPhase3(CPhase & p);

void Driver::scoreWithoutParents(vector<CPhase> & P, 
				 CColVector & K,
				 CMatrix & B,
				 CMatrix & W,
				 int & nphase,
				 int nhap,
				 CIndividual * proband,
				 int unif)
{

    if (par::debug) 
	cerr << "scoreWithoutParents()" << proband->getID() << "\n";
    

    // Get number of phases
    int n = P.size();
    
    if (par::dump)
    {
	for (int i=0; i < P.size(); i++)
	{
	    cout << "\n\t kid = " ;
	    displayPhase3(P[i]);
	}
	cout << "\n";
    }

    // Vector K of phase probabilities
    K.setCol(n);
    
    // Matrices B and W: scoring matrices
    B.setRow(n);
    B.setCol(nhap);
    B.clear();
    
    W.setRow(n);
    W.setCol(nhap);
    W.clear();

    nphase = n;

    //////////////////
    // Populate K 

    // For unconditional analysis
    if (!par::cond)
	for (int p=1; p <= n; p++)
	    K[p] = P[p-1].prob;
    // For conditional analysis 
    else	
	for (int p=1; p <= n; p++)
	{
	    
	    map<CHaplotype,int>::iterator patit 
		= ind::hcode.find(P[p-1].paternal);
	    
	    map<CHaplotype,int>::iterator matit 
		= ind::hcode.find(P[p-1].maternal);
	    
	    // should never occur...
	    if ( patit == ind::hcode.end() ||
		 matit == ind::hcode.end() )
		error("Internal error -- haplotype gone missing");
	    
	    int patn = patit->second;
            int matn = matit->second;
	    
	    if (par::dump)
		cout << "\n element " << p << " = " 
		     << ind::hapfreq1[patn] << " * " << ind::hapfreq1[matn]
		     << "  = " << ind::hapfreq1[patn] * ind::hapfreq1[matn]  ;
	    
	    K[p] = ind::hapfreq1[patn] * ind::hapfreq1[matn] ;

	}
    

    // Populate B (unrelateds have no W component)
    // Order of haplotypes in B and W is given by 
    // ind::hfreq map (which contains haplotype freqs)
    for (int p=1; p <= n; p++)
    {	    
	map<CHaplotype,int>::iterator patit 
	    = ind::hcode.find(P[p-1].paternal);
	
	map<CHaplotype,int>::iterator matit 
	    = ind::hcode.find(P[p-1].maternal);
	
	// should never occur...
	if ( patit == ind::hcode.end() ||
	     matit == ind::hcode.end() )
	    error("Internal error -- oh dear, a haplotype has gone missing...");
	
	int patn = patit->second;
	int matn = matit->second;
	
	// score diploid
	if (!par::xchr || (par::xchr && proband->sex ==2 ) )
	{
	  // genetic models 
	  if (par::model_dom) // Dominant
	    {
	      B[p][patn] = B[p][matn] = 1;
	    }
	  else if (par::model_rec) // Recessive
	    {
	      if (patn == matn)
		B[p][patn] = 1;
	    }
	  else // Multiplicative
	    {
	      B[p][patn]++;
	      B[p][matn]++;	    
	    }
	}
	else // male X chromosome : nb, matn & patn will be identical...
	{
	    B[p][matn]++;
	}
    }
    
    // Collapse ambiguous phases if --noweight option is in effect
    if (!par::weight)
    {
	// generate temporary copies of B, W and K 
	CMatrix Bt = B;
	CMatrix Wt = W;
	
	// resize
	B.setRow(1);
	B.clear();
	W.setRow(1);
	W.clear();
	
	for (int h=1; h<=nhap; h++)
	    for (int p=1; p <= K.getCol(); p++)
	    {
		B[1][h] += K[p] * Bt[p][h];
		W[1][h] += K[p] * Wt[p][h];
	    }	
	
	// Only a single phase remaining
	K.setCol(1);
	K[1] = 1;
	nphase = 1;
    }


    if (par::dump)
    {
	cout << "\nPRE-COLLAPSE";
	K.display();
	cout << "\nSum of K = " << K.sum() << "\n";
	cout << "matrix B";
	B.display();
	cout << "\n matrix W";
	W.display();
	cout << "\n ------------------------------------------- \n";
    }


    // Collapse phase speedup
    if (par::collapse_phase_speedup)
    {
	helper::collapsePhases(K,B,W);
	nphase = K.getCol();
    }
    
    // Remove if uniformative (too many phases)
    // NOTE :: when scoring() is repeated this will
    // need to be done earlier... 
    if (nphase >= unif && unif > 0)
    {
	P.clear();
	K.setCol(0);
	B.setRow(0);
	W.setRow(0);
	nphase = 0;
    }
    
    
    // Standardise...
    double sumK = proband->box->K.sum();
    if (sumK>0)
	proband->box->K.multiply(1/sumK);
	   
    
    if (par::dump)
    {
	cout << "\nPOST-COLLAPSE";
	K.display();
	cout << "\nSum of K = " << K.sum() << "\n";
	cout << "matrix B";
	B.display();
	cout << "\nmatrix W";
	W.display();
	cout << "\n ------------------------------------------- \n";
    }
    
}

void Driver::scoreWithParents(vector<CParentalPhase> & PP, 
			      CColVector & K,
			      CMatrix & B,
			      CMatrix & W,
			      int & nphase,
			      int nhap,
			      CIndividual * proband,
			      int unif)
{

    if (par::debug) 
	cerr << "scoreWithParents() for " << proband->getID() <<" \n";
    
    
    // Get number of phases
    int npp = PP.size();
    int n = 0;
    for (int pp=0; pp < npp; pp++)
	n += PP[pp].P.size();
    
    // Vector K of phase probabilities
    K.setCol(n);
    
    // Matrices B and W: scoring matrices
    B.setRow(n);
    B.setCol(nhap);
    B.clear();
    
    W.setRow(n);
    W.setCol(nhap);
    W.clear();
    
    nphase = n;

    int cnt = 1;
    
    
    ////////////////////////////////
    // Populate B and W for trios
    /////////////////////////////////
    // Consider each Parental Phase
    for (int pp=0; pp < npp; pp++)
    {

	if (par::debug) cerr << "considering parental phase " << pp << " of " << npp << "\n";

	if (par::dump)
	{
	    cout <<"\n\n paternal = " ;
	    displayPhase3(PP[pp].father);
	    cout <<"\t maternal = " ;
	    displayPhase3(PP[pp].mother);
	    for (int i=0; i < PP[pp].P.size(); i++)
	    {
		cout << "\n\t kid = " ;
		displayPhase3(PP[pp].P[i]);
	    }
	    cout << "\n";
	}

	// Get haplotype frequencies for parental haplotypes
	map<CHaplotype,int>::iterator patit1 
	    = ind::hcode.find(PP[pp].father.paternal);
	map<CHaplotype,int>::iterator patit2 
	    = ind::hcode.find(PP[pp].father.maternal);
	map<CHaplotype,int>::iterator matit1
	    = ind::hcode.find(PP[pp].mother.paternal);
	map<CHaplotype,int>::iterator matit2 
	    = ind::hcode.find(PP[pp].mother.maternal);
	// should never occur
	if ( patit1 == ind::hcode.end() ||
	     patit2 == ind::hcode.end() ||
	     matit1 == ind::hcode.end() ||
	     matit2 == ind::hcode.end() )
	    error("Internal error -- haplotype gone missing");
	// The four haplotype frequencies
	double patfreq1 = ind::hapfreq1[patit1->second];
	double patfreq2 = ind::hapfreq1[patit2->second];
	double matfreq1 = ind::hapfreq1[matit1->second];
	double matfreq2 = ind::hapfreq1[matit2->second];
	
	if (par::dump)
	{
	    cout << " version 1 : (not summed to 1 necessarily...";
	    cout << "\n pat hap  = " << patfreq1 << " * " << patfreq2 << " = " << patfreq1 * patfreq2 ;
	    cout << "\n mat hap  = " << matfreq1 << " * " << matfreq2 << " = " << matfreq1 * matfreq2 ;
	    cout << "\n pp_prob = " << patfreq1 * patfreq2 * matfreq1 * matfreq2 ;
	}
	
	// Fixed parental phase prob
	// (could be calculated under joint estimation)
	
	double pp_prob = patfreq1 * patfreq2 * matfreq1 * matfreq2 * PP[pp].mult; 
	
	
	//////////////////////////////////////////
	// Calculate expected given parents (B)
	
	// all possible phases | parental phase
	vector<CPhase> poss = enumerateGivenParents(PP[pp].father,
						    PP[pp].mother,
						    locmap.theta(),
						    par::recomb);
	
	// Get number of phases | parental phase
	int all_n = poss.size();
	CVector all_K(all_n);
	CMatrix all_B(all_n, nhap);
	
	for (int p=1; p <= all_n; p++)
	{
	    all_K[p] = poss[p-1].prob;
	    
	    map<CHaplotype,int>::iterator patit 
		= ind::hcode.find(poss[p-1].paternal);
	    map<CHaplotype,int>::iterator matit 
		= ind::hcode.find(poss[p-1].maternal);
	    
	    
	    // nb . for now, skip if not on list.... 
	    //      this will have to change to allow for recombination
	    if ( patit != ind::hcode.end() &&
		 matit != ind::hcode.end() )
	    {		
		int patn = patit->second;
		int matn = matit->second;
		
		
		// genetic models: Between 
		if (par::model_dom) // Dominant
		  {
		    all_B[p][patn] = all_B[p][matn] = 1;
		  }
		else if (par::model_rec) // Recessive
		  {
		    if (patn == matn)
		      all_B[p][patn] = 1;
		  }
		else // Multiplicative
		  {
		    all_B[p][patn]++;
		    all_B[p][matn]++;	    
		  }
		
	    }
	}
	
	// Vector of expected scores
	CVector exp_H(nhap);
	for (int p=1; p <= all_n; p++)
	    for (int h=1; h <= nhap; h++)
		exp_H[h] += all_K[p] * all_B[p][h];
	
	// If multiplicative, scale exp_H so that it sums to 2
	// or else sums to 1
	double exp_H_total = exp_H.sum();
	if (par::model_dom || par::model_rec)
	  exp_H.multiply(1/exp_H_total);
	else
	  exp_H.multiply(2/exp_H_total);
	
	//////////////////////////////////
	// Consider each offspring phase
	for (int p=0; p < PP[pp].P.size(); p++)
	{
	    
	    // Calculate W for each haplotype
	    CVector within(nhap);
	    
	    map<CHaplotype,int>::iterator patit 
		= ind::hcode.find(PP[pp].P[p].paternal);
	    
	    map<CHaplotype,int>::iterator matit 
		= ind::hcode.find(PP[pp].P[p].maternal);
	    
	    // should never occur...
	    if ( patit == ind::hcode.end() ||
		 matit == ind::hcode.end() )
		error("Internal error -- haplotype gone missing");
	    
	    int patn = patit->second;
	    int matn = matit->second;
	    
	    
	    // genetic models: Between 
	    if (par::model_dom) // Dominant
	      {
		within[patn] = within[matn] = 1;
	      }
	    else if (par::model_rec) // Recessive
	      {
		if (patn == matn)
		  within[patn] = 1;
	      }
	    else // Multiplicative
	      {
		within[patn]++;         
		within[matn]++;	    
	      }
	    
	    
	    // Populate score matrix B 
	    for (int h=1; h <= nhap; h++)
		B[cnt][h] = exp_H[h];
	    
	    // Populate score matrix W
	    for (int h=1; h <= nhap; h++)
		W[cnt][h] = within[h] - exp_H[h]; 
	    

	    // Populate K
	    if (!par::cond)
		K[cnt] = PP[pp].P[p].prob * pp_prob ;
	    else
	    {
		
		map<CHaplotype,int>::iterator patit 
		    = ind::hcode.find(PP[pp].P[p].paternal);
		
		map<CHaplotype,int>::iterator matit 
		    = ind::hcode.find(PP[pp].P[p].maternal);
		
		// should never occur...
		if ( patit == ind::hcode.end() ||
		     matit == ind::hcode.end() )
		    error("Internal error -- haplotype gone missing");
		
		int patn = patit->second;
		int matn = matit->second;
		
		K[cnt] = PP[pp].P[p].prob * pp_prob ; 
		
	    }
	    
	    // Increment combined phase counter
	    cnt++;
	} // next p
    } // next pp
    
    
    // Collapse ambiguous phases if --noweight option is in effect
    if (!par::weight)
    {
	// generate temporary copies of B, W and K 
	CMatrix Bt = B;
	CMatrix Wt = W;
	
	// resize
	B.setRow(1);
	B.clear();
	W.setRow(1);
	W.clear();
	
	for (int h=1; h<=nhap; h++)
	    for (int p=1; p <= K.getCol(); p++)
	    {
		B[1][h] += K[p] * Bt[p][h];
		W[1][h] += K[p] * Wt[p][h];
	    }	
	
	// Only a single phase remaining
	K.setCol(1);
	K[1] = 1;
	nphase = 1;
    }


    if (par::dump)
    {
	cout << "\nPRE-COLLAPSE";
	K.display();
	cout << "\nSum of K = " << K.sum() << "\n";
	cout << "matrix B";
	B.display();
	cout << "\nmatrix W";
	W.display();
       	cout << "\n-------------------------------------------------------\n";
    }

    // Collapse phase speedup
    if (par::collapse_phase_speedup)
    {
	helper::collapsePhases(K,B,W);
	nphase = K.getCol();
    }
    
    // Remove if uniformative (too many phases)
    // NOTE :: when scoring() is repeated this will
    // need to be done earlier... 
    if (nphase >= unif && unif > 0)
    {
	PP.clear();
     	K.setCol(0);
     	B.setRow(0);
     	W.setRow(0);
     	nphase = 0;
    }
    
    // Standardise...
    double sumK = proband->box->K.sum();
    if (sumK>0)
	proband->box->K.multiply(1/sumK);


    if (par::dump)
    {
	cout << "\nPOST-COLLAPSE";
	K.display();
	cout << "\nSum of K = " << K.sum() << "\n";
	cout << "matrix B";
	B.display();
	cout << "\nmatrix W";
	W.display();
       	cout << "\n-------------------------------------------------------\n";
    }
    

}


vector<CPhase> Driver::allPossibleWithoutParents(int nl, int nh)
{

    vector<CPhase> all_phase;
    
    for (int h1 = 1 ; h1 <= nh ; h1++)
	for (int h2 = 1 ; h2 <= nh ; h2++)
	{
	    CPhase newp(nl);
	    newp.paternal = ind::hap[h1];
	    newp.maternal = ind::hap[h2];
	    newp.prob = ind::hapfreq1[h1] * ind::hapfreq1[h2];
	    all_phase.push_back(newp);
	} 
    
    return all_phase;
    
}





vector<CParentalPhase> Driver::allPossibleWithParents(int nl, int nh)
{
    
    // consider all possible mother, father, use get enumerateGivenParents for kid,
    
    if (par::debug)
	cerr << "\n in allPossibleWithParents() \n";

    vector<CParentalPhase> all_phase;

    // construct list of all possible unique haplotype pairs
    vector<int> ga;
    vector<int> gb;
    for (int g1=1; g1<=nh; g1++)
	for (int g2=g1; g2<=nh; g2++)
	{
	    ga.push_back(g1);
	    gb.push_back(g2);
	}    
    int gn = ga.size();
    
    // assign to parents
    for (int pat=0; pat<gn; pat++)
	for (int mat=pat; mat<gn; mat++)
	{
	    int mult = 1;
	    // paternal het
	    if (ga[pat] != gb[pat]) mult*=2;
	    // maternal het
	    if (ga[mat] != gb[mat]) mult*=2;
	    // paternal 'het'
	    if (mat != pat) mult *=2;
	    
	    CPhase fphase(nl);
	    fphase.prob = ind::hapfreq1[ga[pat]] * ind::hapfreq1[gb[pat]] ;
	    fphase.paternal = ind::hap[ga[pat]];
	    fphase.maternal = ind::hap[gb[pat]];
	    
	    CPhase mphase(nl);
	    mphase.prob = ind::hapfreq1[ga[mat]] * ind::hapfreq1[gb[mat]];
	    mphase.paternal = ind::hap[ga[mat]];
	    mphase.maternal = ind::hap[gb[mat]];
	    
	    CParentalPhase new_phase;
//	    new_phase.prob   = fphase.prob * mphase.prob;
	    new_phase.mult = mult;
	    new_phase.father = fphase;
	    new_phase.mother = mphase;
	    new_phase.P = enumerateGivenParents(fphase, 
						mphase,
						locmap.theta(),
						par::recomb);
	    all_phase.push_back(new_phase);
	    
	    
	}
    
    return all_phase;
    
}


void displayPhase3(CPhase & p)
 {
     int l = p.paternal.pos().size();
     for (int j=1; j<=l; j++)
 	cout << p.paternal.pos(j);
     cout << "\\";
     for (int j=1; j<=l; j++)
 	cout << p.maternal.pos(j);
     cout << "\t" << p.prob;     
     if (p.prob==0) cout << " * ";
 }
